

/* A FAIRE pour la version suivante a ecrire : si  "bipfile"
   n'existe pas (ou mieux: si un flag lui dit qu'il ne marche pas
   apres Q* il ne cherche pas le fichier "bipfile" mais qu'il genere
   automatiquement les bipartitions triviales en memoire avant de les
   passer a tree-pop */ 


/***************************************************************************/
/*  
 
   ALGORITHME inferant un arbre en resolvant progrssivement les noeuds
   de degre > 3 presents dans l'arbre infere par la methode Q*
   (programme qstar).

   Note : on a la propriete que les differents noeuds irresolus de
   l'arbre initial concernent des ensembles de quadruplets disjoints.

   Specifs : on part du fichier "bipfile", et par l'algo
   lineaire de tree-popping on constitue une struct d'arbre en memoire
   representant ||*. Puis on resoud chq noeud de degre > 3  separement
   par le meme algo que "glouton" (complexite en O(n^4)).


ATTENTION : on peut forcer l'algo a inferer un arbre totalement
            resolu. C'est ok si on veut optimiser le critere NMQS 
	    (Nombre MAx de Quadruplets Satisfaits), mais ca
	    n'est pas toujours approprie si on cherche le meilleur
	    estimateur de la phylogenie des taxa etudiees.

*/
/*-------------------------------------------------------------------------*/
/* PARAM = -a (valeur de "a" ds critere de construction : SAT-a.CONTREDITS) */
/* ENTREE = 
     fic "bipfile" : liste des bipartitions de l'arbre a completer.
         Si on desire partir de la topologie etoile (la moins
	 resolue), le fichier "bipfile" doit contenir les bipartitions
	 triviales (un taxon oppose a tous les autres.

     fic "quartfile.left" : liste des quadruplets a resoudre (donc
	 non-resolus par les bipartitions de "bipfile") pour completer
	 l'arbre (doit contenir la correspondance numero-taxa).	 
     fic "quartfile.res" : liste des quadruplets resolus par l'arbre
     que AddQuart complete (doit contenir la correspondance numero-taxa)    */
/* SORTIE = 
    fic "bipfile" complete par les bips induites par l'algo AddQuart        */


/*--------------------------------------------------------------------------*/
/* VOIR si DIST_TOPO accepte un arbre non-binaire ??? */
/* voir pour CONSENSE aussi, ainsi que si il correctement le
  consensus d'arbres enracines pas au meme endroit, ie s'il 
  travaille en non-enracine ou pas ???? */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

FILE *farbre,*fcritere;

#include "resoud-n4.c"



/*-----------------------------------------*/
void Lit_quad_restant(int *nbquads)
/* SE : lit les r4-arbres de Q qui restent encore a resoudre */
/* SI : charge le fichier "quads.1" dans MatQ */
{
FILE *fquad;
char neoquad[100];
int t1,t2,t3,t4,tmp,tmp2,combien;
/* init l'ens de quads */
for (t1=1;t1<=Nb_taxa;t1++)
  for (t2=t1+1;t2<=Nb_taxa;t2++)
    for (t3=t1+1;t3<=Nb_taxa;t3++)
      for (t4=t3+1;t4<=Nb_taxa;t4++)
        MatQ[t1][t2][t3][t4]=0;
if (!(fquad=fopen("quartfile.left","r"))) 
    {printf("ERROR: FILE quartfile.left NOT FOUND\n");fflush(stdout);	/* ++pac */
exit(-1);}

/* Lit le nb de taxa */
fscanf(fquad,"%d",&combien);
if (combien!=Nb_taxa) {printf("ERROR: INCOHERENT NUMBER OF TAXA BETWEEN bipfile AND quartfile.left FILES !!!!!!!\n"); fflush(stdout);	/* ++pac */
exit(-2);}
/* Lit la correspondance numero-espece mais ne la garde pas, on la
   connait deja du fichier "bipfile" */
for (t1=1;t1<=Nb_taxa;t1++)
  fscanf(fquad,"%d %s\n",&tmp,&tmp2);
fscanf(fquad,"\n");

/* Lit les quadruplets */
while (fscanf(fquad,"%s",neoquad)!=EOF)
  {
   /* nouveau quadruplet */
   (*nbquads)++;
   t1=atoi(neoquad);
   fscanf(fquad,"%s",neoquad);
   t2=atoi(neoquad);
   /* ordonne les taxas par ordre croissant */
   if (t1>t2) {tmp=t1;t1=t2;t2=tmp;}
   fscanf(fquad,"%s",neoquad); /* lit le || */
   fscanf(fquad,"%s",neoquad);
   t3=atoi(neoquad);
   fscanf(fquad,"%s",neoquad);
   t4=atoi(neoquad);
   /* ordonne les taxas par ordre croissant */
   if (t3>t4) {tmp=t3;t3=t4;t4=tmp;}
   if (t1<t3) {/*printf("(%d,%d,%d,%d)\n",t1,t2,t3,t4);*/
	       MatQ[t1][t2][t3][t4]=1;}
   else {/*printf("(%d,%d,%d,%d)\n",t3,t4,t1,t2);*/
	 MatQ[t3][t4][t1][t2]=1;}
   /*   if (t1<t3) printf("%d%d || %d%d , ",t1,t2,t3,t4);
   else printf("%d%d || %d%d , ",t3,t4,t1,t2); */
  }
}
/*-------------------------------------------*/
void Lit_bipartitions()
{
FILE *f;
char s[11];
int numbip=1,n=0,partie=1,nbobj=0;
int i,k;
/* RAZ les bips */
for (i=1;i<=Max_nb_noeuds;i++) for(k=0;k<Max_nb_taxa;k++)
  {bipartition[i][1][k]=0;bipartition[i][2][k]=0;}
/* Les charge depuis fichier */
if (!(f=fopen("bipfile","r"))) 
    {printf("ERROR: FILE bipfile NOT FOUND\n");fflush(stdout);	/* ++pac */
exit(-1);}

/* Lit le nb de taxa */
fscanf(f,"%d",&Nb_taxa);
printf("%d taxa\n",Nb_taxa);
if (Nb_taxa>50) {printf("Sorry, this program can't handle more than 50 taxa.\n");exit(0);}
/* Lit la correspondance numero-espece */
for (i=1;i<=Nb_taxa;i++)
  fscanf(f,"%d %s\n",&k,&name[i]);
fscanf(f,"\n");

/* Lit les bipartitions */
while (fscanf(f,"%s",s)!=-1)
   {
    /*printf("%s ",s);*/nbobj++;
    if (!strcmp(s,"|")) {partie++;n=0;}
    else {bipartition[numbip][partie][n]=atoi(s);n++;}
    if (nbobj==Nb_taxa+1)
	{/* lit le poids de la bipartition*/
	 fscanf(f,"%s",s);/*printf("le pds de la bip est %s\n",s);*/
         n=0;nbobj=0;numbip++;partie=1;}
   }
Nb_Bips=numbip-1;fclose(f);
}
/*--------------------------------------------*/

/*_________________________________________________*/
/*_______________  algo de tree-popping ___________*/
/*_________________________________________________*/

void Profondeur(Reste,Atteindre,marques,ind_pile,Pile)
int *Reste,Atteindre[Max_nb_taxa+1],marques[Max_nb_noeuds+1],*ind_pile;
tarbre *Pile[Max_nb_noeuds];
{
/* Parcours l'arbre en profondeur en gerant la pile pr le marquage */
int n,i,v=0;
tarbre *s,*som;
/* Si contient des taxa a atteindre marque les noeuds ds la pile
   depuis le dernier marque de la pile */

som=Pile[(*ind_pile)-1];

for (i=0;som->taxa[i];i++)
  if (Atteindre[som->taxa[i]])
    {
     (*Reste)--;
     n=(*ind_pile)-1; /* marque depuis le sommet jusqu'au premier som
			 marque recontre ds la pile en la descendant */
     while (!marques[Pile[n]->num]) 
       { 
         marques[Pile[n]->num]=1;
         n--;}
    }

/* Parcours ses fils tant que certains taxa restent a joindre */
while ((*Reste>0) && (som->voisins[v]))
  {
   if (((*ind_pile)>1) && (Pile[(*ind_pile)-2]==som->voisins[v]) )
       {;/*printf(" son premier voisin est son pere ds la pile\n");*/}
   else
        {
	 s=som->voisins[v];
	 Pile[*ind_pile]=s;(*ind_pile)++; /* empile s */
	 Profondeur(Reste,Atteindre,marques,ind_pile,Pile); 
	 (*ind_pile)--;Pile[*ind_pile]=NULL; /* depile s */
         }
   v++;
  }
}
/*--------------------------------------------*/
void Parcours(b,partie,marques)
int b,partie,marques[Max_nb_noeuds+1];
/* Met dans "marques" le nom de tous les sommets de l'Arbre
   correspondant au sous-arbre d'une des 2 parties de la bipartition b */
/* Pour ca effectue un parcours en profondeur depuis un sommet de
   cette partie */
/* Comme la topologie courante est un arbre on a la propriete que les
   voisins d'un sommet issu du noeud i, ne sont pas deja atteints,
   sauf le voisin qui correspond au noeud i justement. */
/* Pile contient ds une dim le num des noeuds et a cote si ils sont
   marques(1) ou pas(0) */
{
int Atteindre[Max_nb_taxa+1];
tarbre *Pile[Max_nb_noeuds];
int ind_pile,Reste_a_Atteindre=0;
int i,taxon_dep;
/*---- Init ---- */
    /* cree le vecteur de taxa a atteindre pour cet partie -> O(1)
       pour savoir plus tard si un noeud donne doit etre marque ou pas */
    for (i=1;i<=Max_nb_taxa;i++) Atteindre[i]=0;
    for (i=0;bipartition[b][partie][i]!=0;i++) 
      {Atteindre[bipartition[b][partie][i]]=1;Reste_a_Atteindre++;}
   /* Init la pile a vide */
   for (i=0;i<Max_nb_noeuds;i++) Pile[i]=NULL;
/*---------------*/

/* On part d'un sommet de Arbre qui est un des som qu'on veut atteindre */
taxon_dep=bipartition[b][partie][0]; /* non vide */
/*printf("depart taxon %d donc du noeud  %d\n",taxon_dep,sommet[taxon_dep]->num);*/
ind_pile=0;
Pile[ind_pile]=sommet[taxon_dep];
marques[sommet[taxon_dep]->num]=1;
ind_pile++;

/* parcours proprement dit */
Profondeur(&Reste_a_Atteindre,Atteindre,marques,&ind_pile,Pile);

}
/*--------------------------------------------*/
int Intersection(gche,dte,noeud)
int gche[Max_nb_noeuds+1],dte[Max_nb_noeuds+1],*noeud;
{
int nb=0,i;
for (i=1;i<=Max_nb_noeuds;i++)
 if ((gche[i]) && (dte[i])) {nb++;*noeud=i;}
return(nb);}
/*--------------------------------------------*/
int Appartient(b,part,taxon)
int b,part,taxon;
{
int trouve=0,t=0;
while ((!trouve) && (bipartition[b][part][t])) 
 {if (bipartition[b][part][t]==taxon) trouve=1;
  t++;}
return(trouve);
}
/*--------------------------------------------*/
void Scinde(nd,b,arbG,arbD)
int nd,b,arbG[Max_nb_noeuds+1],arbD[Max_nb_noeuds+1];
{
/* separe le noeud en 2 et repartit les voisins de l'ancien noeud de
   facon a ce que les sous-arbres associes a chq partie de la
   bipartition en question soient connexes */
tarbre *neosom,*noeud,*vois;
int ind,ivieux,nbvois,i,j,v,vvieux;
tarbre *voisin_Dt[Max_nb_noeuds];
int taxa_Dt[Max_nb_taxa];
noeud=noeuds[nd];
if ((neosom=(tarbre *)malloc(sizeof(tarbre)))==0)
     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
     exit(-3);}
neosom->num=Num_prochain_noeud;Num_prochain_noeud++;noeuds[neosom->num]=neosom;
/* lui associe les TAXA de l'ancien appartenant a la partie gche de la
   bipartition : O(n) */
ind=0;ivieux=0;
for (i=0;noeud->taxa[i];i++)
  {
   if (Appartient(b,1,noeud->taxa[i]))
     {neosom->taxa[ind]=noeud->taxa[i];ind++;
      sommet[noeud->taxa[i]]=neosom;}
   else {taxa_Dt[ivieux]=noeud->taxa[i];ivieux++;}
  }

for (j=ind;j<Max_nb_taxa;j++) neosom->taxa[j]=0;           
/* a l'ancien laisse les taxa de la partie dte */
for (i=0;i<ivieux;i++) 
  {noeud->taxa[i] = taxa_Dt[i];}
for (j=i;j<Max_nb_taxa;j++) noeud->taxa[j]=0;

/* lui associe les voisins de l'ancien sommet qui appartiennent au
   sous-arbre associe a la partie gche de la bipartition. Les voisins
   de l'autre sousarbre les met ds voisin_Dt : O(n) */
v=0;
vvieux=0;nbvois=0;
while (noeud->voisins[vvieux])
  {
   if (arbG[noeud->voisins[vvieux]->num])
      {neosom->voisins[v]=noeud->voisins[vvieux];v++;}
   else {voisin_Dt[nbvois]=noeud->voisins[vvieux];nbvois++;} 
   vvieux++;
   }
neosom->voisins[v]=NULL;
/* chge le pointeur des noeuds VOISINS vers le neo noeud : O(n) */
for (v=0;neosom->voisins[v];v++)
  {
   vois=neosom->voisins[v];
   for (i=0;(vois->voisins[i]) && (vois->voisins[i]!=noeud);i++)   ;
   if (!vois->voisins[i]) 
     {printf("ERREUR : je ne trouve pas l'ancien noeud ds la liste des");
      printf("voisins de ses voisins\n");exit(-1);}
   vois->voisins[i]=neosom;
  }

/* garde que les voisins de la sousarb dte dans l'ancien noeud */
for (i=0;i<nbvois;i++) noeud->voisins[i]=voisin_Dt[i];
for (i=nbvois;i<Max_nb_noeuds;i++) noeud->voisins[i]=NULL;

/* ajoute une arete entre le neo et l'ancien noeud */
noeud->voisins[nbvois]=neosom;
neosom->voisins[v]=noeud;v++;neosom->voisins[v]=NULL;

/*printf("les noeuds voisins du neo som sont : ");
for (i=0;neosom->voisins[i];i++) printf("%.2d ",neosom->voisins[i]->num);
printf("\n"); printf("les noeuds voisins de l'ancien som sont : ");
for (i=0;noeud->voisins[i];i++) printf("%.2d ",noeud->voisins[i]->num);
printf("\n");printf("Taxa(noeud) = ");
for (i=1;sommet[i];i++) printf("%.2d(%d) ",i,sommet[i]->num); printf("\n");*/
}

/*_________________________________________________*/
/*____________ fin de l'algo tree-popping _________*/
/*_________________________________________________*/

/*--------------------------------------------*/
void Integre_bipartition(b)
int b;
/* Traite la bieme bipartition */
{
int noeud=-1;
int parcours_gche[Max_nb_noeuds+1],parcours_dte[Max_nb_noeuds+1]; 
int i;

/* si il s'agit de LA bipartition LA PLUS triviale ne rien faire */
/* (laisse passer les aretes externes !) */
if ((bipartition[b][1][0]==0) || (bipartition[b][2][0]==0)) 
   {/*printf("il s'agit d'une bipartition vide\n");*/return;}
/* Init */
for (i=0;i<=Max_nb_noeuds;i++) {parcours_gche[i]=0;parcours_dte[i]=0;}
/* Cherche noeuds de partie gche */
Parcours(b,1,parcours_gche);
/*printf("Sous-arbre gche = noeuds ");
for (i=1;i<=Max_nb_noeuds;i++) 
  if (parcours_gche[i]) printf("%.2d ",i);
printf("\n");*/

/* Cherche noeuds de partie dte */
Parcours(b,2,parcours_dte);
/*printf("Sous-arbre dt = noeuds ");
for (i=1;i<=Max_nb_noeuds;i++) 
  if (parcours_dte[i]) printf("%.2d ",i);
printf("\n");*/
/* Si intersection == 2 noeuds -> incompatibilite 
                   == 0 noeuds -> rien faire 
		   == 1 noeud -> le separer en 2 */

switch (Intersection(parcours_gche,parcours_dte,&noeud))
	{
	case 0:return;break;
	case 2:{printf("ERROR: INCOMPATIBLE BIPARTITIONS !!!!\n");exit(-1);}break;
	case 1:{/* recois le "noeud" comme seule intersection ;*/
		Scinde(noeud,b,parcours_gche,parcours_dte);}
        }
}
/*--------------------------------------------*/
void Affiche_voisins()
{
int i,j;
for (i=1;noeuds[i];i++)
  {printf("n%d (%.2d) -> ",noeuds[i]->num,noeuds[i]->taxa[0]);
   for (j=0;noeuds[i]->voisins[j];j++)
       printf("n%d ",noeuds[i]->voisins[j]->num);
  printf("\n");
  }
}
/*--------------------------------------------------*/
void Affiche_Arbre(noeud,pere)
tarbre *noeud,*pere;
{
int i,prem;
/* Produit un arbre ENRACINE depuis la structure de donnees */

/* affiche ses taxa */
for (i=0;noeud->taxa[i];i++) printf("%.2d:[n%d]",noeud->taxa[i],noeud->num);

/* recurse */
if (noeud->voisins[1]) /* a un autre voisin que son pere */
    {
     printf("(");prem=1;
     for (i=0;noeud->voisins[i];i++)
       {
	  if (noeud->voisins[i]!=pere)  
	    {if (!prem) printf(",");
	     else prem=0;
	     Affiche_Arbre(noeud->voisins[i],noeud);
	    }
	}
     printf("):[n%d]",noeud->num);
    }
}
/*--------------------------------------------------*/
int Aff_Agi()
{
int i,j;listequads *q;
/*
for (i=0;i<Nb_noeuds_agreg;i++)
  for (j=0;j<i;j++)
    {printf("ag(n%d,n%d) : ",refinv[i],refinv[j]);
     if (Agi[i][j].Lqd)
       {q=Agi[i][j].Lqd;
	do
	  {printf("%d%d||%d%d ",q->gche[0],q->gche[1],q->dte[0],q->dte[1]);
	   q=q->suiv;}
	while (q!=Agi[i][j].Lqd);} 
     printf("\n");}
*/
printf("\n(x,y,w) des Agi:\n");
printf("  ");
for (j=0;j<Nb_noeuds_agreg;j++) printf("    %d    ",j);printf("\n");
for (i=0;i<Nb_noeuds_agreg;i++)
  {printf("%d ",i);
   for (j=0;j<i;j++)
     printf(" (%d,%d,%d) ",Agi[i][j].x,Agi[i][j].y,Agi[i][j].w);
 printf("\n");
 }
}
/*-------------------------------*/
void Quadruplets_multi(num)
int num;
{
listequads *qd[2][3],*elem,*prec,*maillon;
int nbconc1=0,nbconc2=0,b1,b2,b3,b4,app,tmp,d,i,lc1,lc2;
int q[3][4]; /* bon ordonnancement des tax pour chacun des 3 r4-arbres */
Ttaxon *l1,*l2,*l3,*l4,*e1,*e2,*e3,*e4;
tarbre *multi;

Correspond_nds_ligcol(num);

multi=noeuds[Multifourches[num]];
/* Pour chq combinaison de 4 branches autour de la multi Mi */
for (b1=0,l1=multi->tax_sousarb[b1];multi->voisins[b1+3];b1++,l1=multi->tax_sousarb[b1])
 for (b2=b1+1,l2=multi->tax_sousarb[b2];multi->voisins[b2+2];b2++,l2=multi->tax_sousarb[b2])
  for(b3=b2+1,l3=multi->tax_sousarb[b3];multi->voisins[b3+1];b3++,l3=multi->tax_sousarb[b3])
     for(b4=b3+1,l4=multi->tax_sousarb[b4];multi->voisins[b4];b4++,l4=multi->tax_sousarb[b4])
    /* pr chq combin de tax appartenant chacun a une des 4 branche choisies */
       for (e1=l1->suiv;e1;e1=e1->suiv)
	 for (e2=l2->suiv;e2;e2=e2->suiv)
	   for (e3=l3->suiv;e3;e3=e3->suiv)
	     for (e4=l4->suiv;e4;e4=e4->suiv)
	       {/*printf("%d%d||%d%d ",e1->num,e2->num,e3->num,e4->num);*/
		/* ordonne correctement les tax des 3 r4-arbres */
		 /* 1er r4-arbre */
	        if (e1->num<e2->num) {q[0][0]=e1->num;q[0][1]=e2->num;}
		else                 {q[0][0]=e2->num;q[0][1]=e1->num;}
		if (e3->num<e4->num) {q[0][2]=e3->num;q[0][3]=e4->num;}
		else                 {q[0][2]=e4->num;q[0][3]=e3->num;}
		if (q[0][0]>q[0][2]) {tmp=q[0][0];q[0][0]=q[0][2];q[0][2]=tmp;
				      tmp=q[0][1];q[0][1]=q[0][3];q[0][3]=tmp;}
		/* 2eme r4-arbre */
	        if (e1->num<e3->num) {q[1][0]=e1->num;q[1][1]=e3->num;}
		else                 {q[1][0]=e3->num;q[1][1]=e1->num;}
		if (e2->num<e4->num) {q[1][2]=e2->num;q[1][3]=e4->num;}
		else                 {q[1][2]=e4->num;q[1][3]=e2->num;}
		if (q[1][0]>q[1][2]) {tmp=q[1][0];q[1][0]=q[1][2];q[1][2]=tmp;
				      tmp=q[1][1];q[1][1]=q[1][3];q[1][3]=tmp;}
		/* 3eme r4-arbre */
	        if (e1->num<e4->num) {q[2][0]=e1->num;q[2][1]=e4->num;}
		else                 {q[2][0]=e4->num;q[2][1]=e1->num;}
		if (e2->num<e3->num) {q[2][2]=e2->num;q[2][3]=e3->num;}
		else                 {q[2][2]=e3->num;q[2][3]=e2->num;}
		if (q[2][0]>q[2][2]) {tmp=q[2][0];q[2][0]=q[2][2];q[2][2]=tmp;
				      tmp=q[2][1];q[2][1]=q[2][3];q[2][3]=tmp;}
		/* Est-ce qu'une des 3 r4-arbres lies a ce quad est ds Q ? */
		app=0;
		if (((MatQ[q[0][0]][q[0][1]][q[0][2]][q[0][3]]) 
		  || (MatQ[q[1][0]][q[1][1]][q[1][2]][q[1][3]]))
		  || (MatQ[q[2][0]][q[2][1]][q[2][2]][q[2][3]])) app=1;
		/* Cree les 6 r4-arbres */
		for (d=0;d<2;d++)
		  for (i=0;i<3;i++)
		  {
		    if ((qd[d][i]=(listequads *)malloc(sizeof(listequads)))==0)
		       {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
		        exit(-3);}
		    if (!d)
		      {qd[d][i]->gche[0]=q[i][0];qd[d][i]->gche[1]=q[i][1];
		       qd[d][i]->dte[0]=q[i][2];qd[d][i]->dte[1]=q[i][3];}
		    else 
		      {qd[d][i]->dte[0]=q[i][0];qd[d][i]->dte[1]=q[i][1];
		       qd[d][i]->gche[0]=q[i][2];qd[d][i]->gche[1]=q[i][3];}
		    /* rattache chacun a son Agi */
		    Quelle_Agi(qd[d][i],&lc1,&lc2);
		    if (Agi[lc1][lc2].Lqd)
		      {qd[d][i]->prec=Agi[lc1][lc2].Lqd->prec;
		       qd[d][i]->prec->suiv=qd[d][i];
		       Agi[lc1][lc2].Lqd->prec=qd[d][i];
		       qd[d][i]->suiv=Agi[lc1][lc2].Lqd;}
		    else {
		      qd[d][i]->prec=qd[d][i]->suiv=qd[d][i];
		      Agi[lc1][lc2].Lqd=qd[d][i];}
		    /* maj le score de cette Agi */
		    if (!app) {qd[d][i]->typ='w';
			       (Agi[lc1][lc2].w)++;}
		    else 
		      {if (MatQ[q[i][0]][q[i][1]][q[i][2]][q[i][3]])
			{(Agi[lc1][lc2].x)++; qd[d][i]->typ='x';}
		      else {(Agi[lc1][lc2].y)++;qd[d][i]->typ='y';}}
		  }
		/* les lie entre eux */
		qd[0][0]->qdsuiv=qd[0][1];qd[0][1]->qdsuiv=qd[0][2];
		qd[0][2]->qdsuiv=qd[1][0];qd[1][0]->qdsuiv=qd[1][1];
		qd[1][1]->qdsuiv=qd[1][2];qd[1][2]->qdsuiv=qd[0][0];
	      }
}
/*-------------------------------*/
int Compte_sat_reletoile()
{
FILE *fq;int nb1=0,tmp,combien;
char s[50];
int t1,t2,t3,t4;
if (! (fq=fopen("quartfile.res","r")) ) 
   return(0); /* ce fichier n'existe pas */

/* Lit le nb de taxa */
fscanf(fq,"%d",&combien);
if (combien!=Nb_taxa) {printf("ERROR: INCOHERENT NUMBER OF TAXA BETWEEN bipfile AND quartfile.res FILES !!!!!!!\n"); exit(-2);}
/* Lit la correspondance numero-espece */
for (t1=1;t1<=Nb_taxa;t1++)
  fscanf(fq,"%d %s\n",&tmp,&s);
fscanf(fq,"\n");
/* Lit les quadruplets */
while (fscanf(fq,"%d %d || %d %d",&t1,&t2,&t3,&t4)!= -1) nb1++;
fclose(fq);
return(nb1);
}
/* -------------------------------------------- */
void Aff_liste(deb)
Ttaxon *deb;
{Ttaxon *p;
for (p=deb->suiv;p;p=p->suiv)
{printf(" %d ",p->num);}
printf("\n");
}

/* -------------------------------------------------- */
void Taxa_Sous_arbo(pere,nd,listaxa)
tarbre *pere,*nd; Ttaxon *listaxa;
/* Determine pour chq noeud (sauf les externes) , quelles sont les
taxa contenus dans la sous-arbo induite par chacun de ses voisins.
Renvoie l'union (ie, les taxa de la sousarbo nd) dans "listaxa"*/
/* SI : listaxa contient un bidon pour pouvoir recuperer le res */
{
int t,fils,indpere,contient_pere[Max_nb_taxa];
Ttaxon *lfils,*neo;
/* Est-ce qu'on est sur un noeud externe (un seul voisin: son pere)*/
if (!nd->voisins[1]) /* alors renvoie le taxon */
  {/*printf("je suis sur un noeud externe : n%d\n",nd->num);*/
   nd->tax_sousarb[0]=NULL;
   if ((neo=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
      exit(-3);}
   neo->num=nd->taxa[0]; /* on sait que now un seul taxa */
   neo->suiv=listaxa->suiv;listaxa->suiv=neo;}
else /* recupere les taxa de chq sousarbo associee a un fils */
  {/*printf("je suis sur un noeud interne : n%d\n",nd->num);*/
    for (t=1;t<=Nb_taxa;t++) contient_pere[t]=1;
    /* pr chq sousarbo fille */
    for (fils=0;nd->voisins[fils];fils++)
      if (nd->voisins[fils]!=pere)
	{/*printf("je vais voir son fils : n%d\n",nd->voisins[fils]->num);*/
	 if ((nd->tax_sousarb[fils]=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
	     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
	      exit(-3);}
         nd->tax_sousarb[fils]->suiv=NULL;
	 Taxa_Sous_arbo(nd,nd->voisins[fils],nd->tax_sousarb[fils]);
         /*printf("Je reviens avec :");Aff_liste(nd->tax_sousarb[fils]);*/
	 /* Union-recopie avec listaxa + enleve de contient_tax pere */
	 for (lfils=nd->tax_sousarb[fils]->suiv;lfils;lfils=lfils->suiv)
	   {
	    if ((neo=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
	      {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
	       exit(-3);}
	    neo->num=lfils->num;neo->suiv=listaxa->suiv;listaxa->suiv=neo;
	    contient_pere[lfils->num]=0;
	   }
        }
      else indpere=fils;
    /* deduit la sous-arbo du pere par {1..n}-union(sousarbo(fils)) */
    if ((nd->tax_sousarb[indpere]=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
      {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
       exit(-3);}
    nd->tax_sousarb[indpere]->suiv=NULL;
    for (t=1;t<=Nb_taxa;t++) 
      if (contient_pere[t])
	{ if ((neo=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
	     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
	      exit(-3);}
	  neo->num=t;
	  neo->suiv=nd->tax_sousarb[indpere]->suiv;
	  nd->tax_sousarb[indpere]->suiv=neo;
	}
   /*printf("J'en deduit que le pere contient les espces ");
   Aff_liste(nd->tax_sousarb[indpere]);*/
  }
}
/*_______________________________________________________*/
/*--------------------------------------------------*/
void Especes_Sous_arbo(pere,racine,ct)
tarbre *racine,*pere; int *ct;
/* Cherche les esps de la sous-arbo partant de "racine" definie par la
   direction opposee a "pere", et remplit la struct "contient_tax"
   depuis l'indice "ct"*/
{
int fils,tax;
/* ajoute les eventuels taxa de ce noeud */
for (tax=0;racine->taxa[tax];tax++)
  {contient_tax[*ct]=racine->taxa[tax];(*ct)++;}
/* parcours la sous-arbo de racine */
for (fils=0;racine->voisins[fils];fils++)
  if (racine->voisins[fils]!=pere)
    Especes_Sous_arbo(racine,racine->voisins[fils],ct);
}
/*--------------------------------------------------*/
int main(argc, argv)
int argc;char *argv[];
{extern char *optarg;
char c,nomfic[20];
int nbquads=0;
int i,j,k,cpt;
Nb_taxa=4;
/* Traite arguments */
while ((c = getopt(argc, argv, "a:")) != -1)
     switch (c) {
     case 'a':a=atof(optarg);break;
     case '?':{printf("PARAMETER ERROR !\n");exit(1);}
                 }
Lit_bipartitions();

#ifdef _PLAN9
srand(getpid()); 
#else
srandom(getpid()); 
#endif

/*--- Init arbre avec ts les taxa ds un noeud --*/
if ((Arbre=(tarbre *)malloc(sizeof(tarbre)))==0)
     {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
     exit(-3);}
for (i=0;i<Nb_taxa;i++) {Arbre->taxa[i]=i+1;sommet[i+1]=Arbre;}
for (i=Nb_taxa;i<Max_nb_taxa;i++) {Arbre->taxa[i]=0;sommet[i+1]=NULL;}
for (i=1;i<=Max_nb_noeuds;i++) noeuds[i]=NULL;
Arbre->num=1;Num_prochain_noeud=2;noeuds[1]=Arbre;
Arbre->voisins[0]=NULL;
/*-----------*/
Critere_global.x=0;
Critere_global.y=0;
Critere_global.w=0;
/*--------------------------------------------------------------*/
/* 1 : creer l'arbre depuis les bipartitions ||* : tree-popping */
/*--------------------------------------------------------------*/
for (i=1;i<=Nb_Bips;i++) Integre_bipartition(i);

/* calcule la list des taxa par sous-arbre par parcours recursif
   depuis le taxon 1 */
if ((sommet[1]->tax_sousarb[0]=(Ttaxon *)malloc(sizeof(Ttaxon)))==0)
  {printf("MEMORY ALLOCATION ERROR (no more space available)\n");
   exit(-3);}
sommet[1]->tax_sousarb[0]->suiv=NULL;

/*printf("Arbre ||* = (01,"); 
Affiche_Arbre(sommet[1]->voisins[0],sommet[1]);printf(");\n");*/

Taxa_Sous_arbo(sommet[1],sommet[1]->voisins[0],sommet[1]->tax_sousarb[0]);   
nbquads=Compte_sat_reletoile();
Critere_global.x+=nbquads;
/*-------------------------------------------------------*/
/* 2 : etablit la liste des multifourches : O(n) 
       se fait simplement en parcourant la struct noeuds */
/*-------------------------------------------------------*/
/* init a vide */
for (i=0;i<Max_nb_noeuds;i++) 
   Multifourches[i]=0;
i=1;
while (noeuds[i])
  {
    if ( ((noeuds[i]->voisins[0])&&(noeuds[i]->voisins[1]))
       && ((noeuds[i]->voisins[2])&&(noeuds[i]->voisins[3])) )
      {Multifourches[nb_multi]=i;nb_multi++;}
    i++;
  }
/*------------------------------------------------------------*/
/* 3 : Lit les quadruplets non-resolus par ||* : "quads.left" */
/*------------------------------------------------------------*/

Lit_quad_restant(&nbquads); /* les place dans le tabl a 4 dim MatQ */
fbip=fopen("bipfile","a"); /* ajoute bips creees dans bipfile */

/*--------------------------------------------------*/
/* 4 : Resoud chq multifourche tour a tour          */
/*--------------------------------------------------*/
printf("\nResolving multifurcations :\n");
for (i=0;i<nb_multi;i++)
  {
   /* Etablir la liste des esps par sous-arbo % multi 
      pas faisable avant car 2 nds multi voisins peuvent ne plus 
      etre voisins apres resolution d'un des 2 */ 
   /*printf("n%d ->  ",Multifourches[i]);*/
   for(j=0;noeuds[Multifourches[i]]->voisins[j];j++)
     {
       for(k=0;k<Nb_taxa;k++) contient_tax[k]=0;
       cpt=0;
       Especes_Sous_arbo(noeuds[Multifourches[i]],noeuds[Multifourches[i]]->voisins[j],&cpt);
     }
    Quadruplets_multi(i);
    /* FAIT L'ALGO "GLOUTON" POUR CETTE MULTIFOURCHE */
    Resoud_multi(i);
 }
fclose(fbip);

/* On obtient ainsi un arbre pas forcement resolu et non-enracine, */
/* qu'on se contente d'afficher pour le moment */
/*printf("Arbre apres resolution des multifourches par glouton = \n");
printf( "(01," );
Affiche_Arbre(sommet[1]->voisins[0],sommet[1]);printf(");\n");*/

printf("\nThe inferred tree satisfies %d (%.2f%%) input quartets\n",Critere_global.x,100.0*(float)Critere_global.x/(float)nbquads);
printf("contradicts %d (%.2f%%) input quartets\n",Critere_global.y,100.0*(float)Critere_global.y/(float)nbquads); 
printf("and resolves %d (%.2f%%) quartets unresolved in the input file\n",Critere_global.w,100.0*(float)Critere_global.w/(float)nbquads);
}






