/* 
   PROG DEDUISANT LES QUADRUPLETS DEPUIS UNE MATRICE DE DISTANCES.

   La MATRICE DE DISTANCE EST LUE POUR L'INSTANT DEPUIS "outfile"
   PRODUIT PAR LE PROG "dnadist" DE PHYLIP QUI CORRIGE LES DATA
   PAR UN MODELE DE KIMURA A DEUX PARAMETRES.

   ON POURRAIT AUSSI CALCULER LES DISTANCE DIRECTEMENT DEPUIS LA 
   MATRICE DE CARACTERES, MAIS PBM : JE NE TROUVE PAS LES MEMES
   DISTANCE CORRIGEES QUE LUI.

   LES QUADRUPLETS SONT INFERES PAR LA METHODE DES DISTANCES : 
      calcule la distance entre tout couple d'especes, 
      puis pour chq quadruplet d'especes infere le quadruplet
      correspondant a la plus petite des 3 sommes de distances.


--------------------------------------------
Parameters: 
    -d : read a distance matrix (1)  or compute it from the sequences (0).
         (default is 0). 
    -l : the distance matrix to be read is lower triangular (1) 
         or squared (0) (default is 0).
    -c : correct (1 or 2) or not (0) the distances according to an 
         evolutionary model (default is 0).
         -c1 correct according to the Jukes-Cantor 1969 model.
         -c2 correct according to the 2-param Kimura 1980 model.
    -w : transV/transI weight applied during distance correction in the 
         2-param Kimura 1980 model.
    -s : save the distance matrix (1) or don't (0) (default is 0)
    -g : (1) use gap position whenever is possible or (0) drop them 
         from the whole analysis (default is 0).
Input : 
 - file "infile.dist" : distance matrix
ou file "infile.nuc" : nucleotide matrix
Output :
 - file "quartfile"
 - file "outfile.dist" : distance matrix 

/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Max_nb_taxa 300
#define Max_car 12001

char Matrice_car[Max_car][Max_nb_taxa+1]; /* !! mat[i][0] pas utilise */
char gap_position[Max_car];

float Matrice_Dist[Max_nb_taxa+1][Max_nb_taxa+1]; 
/* convention [lig][col] et 1/2 matrice inf utilisee : lig > col */

int TransV[Max_nb_taxa+1][Max_nb_taxa+1];
int TransI[Max_nb_taxa+1][Max_nb_taxa+1];
int NbPos[Max_nb_taxa+1][Max_nb_taxa+1];
int Nb_taxa,Nb_cars;
int lower_triang=0;
int use_gap_pos=0; /* default is that gap position are unused */
float Poids_Transv = 2.0; /* weight of transV % transI (ex : 2.0)*/

char nom[Max_nb_taxa+1][20];

/*-----------------------------------------------------------*/
void Lit_Distances()
/* recupere les distances calculees par "dnadist" de phylip */
{
FILE *fd;
float d;
int i,j,num,stop;
if (!(fd=fopen("infile.dist","r"))) 
    {printf("ERROR: FILE infile.dist NOT FOUND\n");exit(-1);}

if (lower_triang) printf("reading lower triangular matrix of distances\n");
else printf("reading squared matrix of distances\n");
fscanf(fd,"%d",&Nb_taxa);
printf("Input size: %d taxa\n",Nb_taxa);
/* Init matrices now that we know Nb_taxa */
for (i=1;i<=Nb_taxa;i++) 
    for (j=1;j<=Nb_taxa;j++) NbPos[i][j]=1;
/* Read taxa names and inter-taxa distances */
for (i=1;i<=Nb_taxa;i++)
  {
  fscanf(fd,"%s",&nom[i]);
  if (lower_triang) stop=i; else stop=Nb_taxa+1;
  for (j=1;j<stop;j++)
    {
     fscanf(fd,"%f",&d);Matrice_Dist[i][j]=d;
    }
  }
}
/*--------------------------------------------------*/
void Sauve_Distances()
{
/* Sauver la matrice de distance ds un fichier comme DNADIST */
FILE *fdist;
int i,j,err=0,stop;
fdist=fopen("outfile.dist","w");
fprintf(fdist,"   %d\n",Nb_taxa);
for (i=1;i<=Nb_taxa;i++)
  {
   fprintf(fdist,"%s",nom[i]);
   for (j=0;j<=10-strlen(nom[i]);j++) fprintf(fdist," ");
   if (lower_triang) stop=i; else stop=Nb_taxa+1;
   for (j=1;j<stop;j++) 
    {  
     if (i>j) 
      {if (NbPos[i][j]) fprintf(fdist,"%.4f  ",Matrice_Dist[i][j]);
        else {fprintf(fdist,"??????  ");err=1;}}
      else if (i<j)
      {if (NbPos[j][i]) fprintf(fdist,"%.4f  ",Matrice_Dist[j][i]);
       else {fprintf(fdist,"??????  ");err=1;}}
      else fprintf(fdist,"0.0000  ");
    }
   fprintf(fdist,"\n");
  }
fclose(fdist);
printf("Inter-taxa distances output in file : outfile.dist\n");
if (err) printf("WARNING: some distances could not be computed due to too much gaps or unknown characters between some sequences!!!!\n");
}
/*--------------------------------------------------*/
void Lit_Matrice_Car()
{
FILE *fmat;
int c,e,i,j;
char car;

/* Read taxa and character numbers */
if (!(fmat=fopen("infile.nuc","r"))) 
    {printf("ERROR: FILE infile.nuc NOT FOUND\n");exit(-1);}
fscanf(fmat,"%d",&Nb_taxa);
fscanf(fmat,"%d",&Nb_cars);
printf("Input size: %d taxa\nSequences of %d characters\n",Nb_taxa,Nb_cars);
fscanf(fmat,"\n");

/* Init matrices and refs now that Nb_taxa is known */
for (i=1;i<=Nb_taxa;i++)
  for (j=1;j<=Nb_taxa;j++)
    {Matrice_Dist[i][j]=0.0;TransV[i][j]=0;TransI[i][j]=0;NbPos[i][j]=0;}

/* initially no position has a gap */
for (c=0;c<Nb_cars;c++) gap_position[c]=0;
/* reading the characters */
for (e=1;e<=Nb_taxa;e++)
  {
   for (i=0;i<10;i++) {fscanf(fmat,"%c",&nom[e][i]);} 
   /*nom[e][i]="\0");*/
   for (c=0;c<Nb_cars;c++)
      {
       fscanf(fmat,"%c",&car);
       switch (car) {
       case 'A':Matrice_car[c][e]='A';break;	 
       case 'a':Matrice_car[c][e]='A';break;
       case 'G':Matrice_car[c][e]='G';break;	
       case 'g':Matrice_car[c][e]='G';break;	
       case 'C':Matrice_car[c][e]='C';break;	
       case 'c':Matrice_car[c][e]='C';break;	
       case 'T':Matrice_car[c][e]='T';break;
       case 't':Matrice_car[c][e]='T';break;
       case 'U':Matrice_car[c][e]='T';break; 
       case 'u':Matrice_car[c][e]='T';break;
       case '?':Matrice_car[c][e]='?';break;
       case '.':{if (e==1) 
                  {printf("ERROR: '.' character in the 1st sequence!!!\n");
		   exit(-1);}
	        else Matrice_car[c][e]=Matrice_car[c][1];};break;
       case '-':{Matrice_car[c][e]='-';
		 gap_position[c]++;}break;
       default: c--; /* ignore space or non-valid chars appearing in seqs */
       }
     }
   fscanf(fmat,"\n");
  }
/*printf("Set of sequences read :\n");
for (e=1;e<=Nb_taxa;e++) 
  { for (c=0;c<Nb_cars;c++)
       printf("%c",Matrice_car[c][e]);
    printf("\n");}*/
fclose(fmat);
}
/*--------------------------------------------------*/
int Transition(c1,c2)
char c1,c2;
{
if ((c1=='A') && (c2=='G')) return(1);
if ((c1=='G') && (c2=='A')) return(1);
if ((c1=='T') && (c2=='C')) return(1);
if ((c1=='C') && (c2=='T')) return(1);
return(0);
}
/*--------------------------------------------------*/
void Affiche_Dist()
{int i,j,stop;

printf("Distance matrix:\n");
for (i=1;i<=Nb_taxa;i++)
 {printf("%s",nom[i]);
  for (j=0;j<=10-strlen(nom[i]);j++) printf(" ");
  if (lower_triang) stop=i; else stop=Nb_taxa+1;
  for (j=1;j<stop;j++) 
    {  
      if (i>j) 
       {if (NbPos[i][j]) printf("%.4f  ",Matrice_Dist[i][j]);
        else printf("??????  ");}
      else if (i<j)
	{if (NbPos[j][i]) printf("%.4f  ",Matrice_Dist[j][i]);
         else printf("??????  ");}
      else printf("0.0000  ");
    }
  printf("\n");
 }
}
/*--------------------------------------------------*/
void Calcule_Distances()
{/* Calcule les distances initiales entres taxa */
int e1,e2,e3,e4,c,egaux,nb_gaps=0;

for (c=0;c<Nb_cars;c++) if (gap_position[c]) nb_gaps++;
if (!nb_gaps) printf("No gap positions.\n"); 
else {printf("gaps at %d positions -- these positions ",nb_gaps);
      if (use_gap_pos)  printf("are however used whenever is possible\n");
      else printf("are dropped from the whole analysis\n");}
/* Compute the lower half matrix of inter-taxa distances */
for (e1=1;e1<=Nb_taxa;e1++)
 for (e2=1;e2<e1;e2++)
  for (c=0;c<Nb_cars;c++)
    {
     if ((Matrice_car[c][e1]!='?') && (Matrice_car[c][e2]!='?'))
      if ((use_gap_pos) || (!gap_position[c]))
        if ((Matrice_car[c][e1]!='-') && (Matrice_car[c][e2]!='-'))
	 {
	   (NbPos[e1][e2])++;
	   if (Matrice_car[c][e1]!=Matrice_car[c][e2])
	     {Matrice_Dist[e1][e2]+=1.0;
	      if (Transition(Matrice_car[c][e1],Matrice_car[c][e2]))
		TransV[e1][e2]++; else TransI[e1][e2]++;}
	 }
    }
/* Ramene les distances aux nb de positions sur lesquelles elles sont
calculees */
for (e1=1;e1<=Nb_taxa;e1++)
 for (e2=1;e2<e1;e2++)
   Matrice_Dist[e1][e2]/=NbPos[e1][e2];

/*printf("Set of sequences read :\n");
for (e1=1;e1<=Nb_taxa;e1++) 
  { for (e2=1;e2<=Nb_taxa;e2++)
       printf("%d ",NbPos[e1][e2]);
    printf("\n");}*/
}
/*--------------------------------------------------*/
void Infere_quads()
/* Infere les quadruplets depuis la matrice de distances. */
/* Chq qud e1e2||e3e4  est ordonne pour avoir e1<e2, e3<e4, e1<e3 */
{
FILE *fquad;
int nbquads=0;
int e1,e2,e3,e4,egal=0;
float d1,d2,d3;
fquad=fopen("quartfile","w");

/* Indique nb de taxa */
fprintf(fquad,"   %d\n",Nb_taxa);
/* Indique correspondance numero de taxa - nom taxa */
for (e1=1;e1<=Nb_taxa;e1++) fprintf(fquad,"%.2d %s\n",e1,nom[e1]);
fprintf(fquad,"\n");

for (e1=1;e1<=Nb_taxa-3;e1++)
 for (e2=e1+1;e2<=Nb_taxa-2;e2++)
  for (e3=e2+1;e3<=Nb_taxa-1;e3++)
   for (e4=e3+1;e4<=Nb_taxa;e4++)
     {
      d1=Matrice_Dist[e2][e1] + Matrice_Dist[e4][e3];
      d2=Matrice_Dist[e3][e1] + Matrice_Dist[e4][e2];
      d3=Matrice_Dist[e4][e1] + Matrice_Dist[e3][e2];
      /* printf("d1 = %.4f , d2 = %.4f , d3 = %.4f\n",d1,d2,d3);  */
      if ((d1<d2) && (d1<d3)) 
	 {fprintf(fquad,"%.2d %.2d || %.2d %.2d\n",e1,e2,e3,e4);nbquads++;}
      else
      if ((d2<d1) && (d2<d3)) 
	 {fprintf(fquad,"%.2d %.2d || %.2d %.2d\n",e1,e3,e2,e4);nbquads++;}
      else
      if ((d3<d1) && (d3<d2)) 
	 {fprintf(fquad,"%.2d %.2d || %.2d %.2d\n",e1,e4,e2,e3);nbquads++;}
      else 
	 {/*EGALITE DES 2 (ou 3) PLUS PETITES*/ egal++;
          /* dans ce cas ne deduit pas de quadruplet */}
     }
fclose(fquad);
printf("%d quartets inferred\n%d quartets leaved unresolved\n",nbquads,egal);
printf("Quartets output in file : quartfile\n");
}
/*--------------------------------------------------*/
void JukesCantor_Correction()
/* Applies the JC69 model with the formula:
   d_JC= -4/3 ln(1-3/4p) 
   p = propotion of differences observed ( (#transI+#transV)/n )
/* "log" here is the natural logarithm : ln */
{
int i,j;
float ti,tv,dcor;
for (i=1;i<=Nb_taxa;i++)
  for (j=1;j<i;j++)
    {
      if ( (1.0-(Matrice_Dist[i][j]*3.0/4.0)<0) ) 
	{printf("ERROR: log of a negative value in the distance correction process (Jukes-Cantor model)\n");exit(-1);}
      Matrice_Dist[i][j] = (-4.0/3.0)*log(1.0-(Matrice_Dist[i][j]*3.0/4.0));
    }
}
/*--------------------------------------------------*/
void Kimura_Correction1()
/* Applies the Kimura 2-parameter model with the formula:
   d_kim = 2(t+2/t)[ -1/4 ln(1-2s/n -v/n) + 1/8 ln(1-2v/n)] 
   t = poids des transversions/transitions
   v = nb of transVertions
   s = nb of transItions
   n = nb of characters */
/* "log" here is the natural logarithm : ln */
{
int i,j;
float ti,tv,dcor;
for (i=1;i<=Nb_taxa;i++)
  for (j=1;j<i;j++)
    { ti=(float)TransI[i][j]/NbPos[i][j]; 
      tv=(float)TransV[i][j]/NbPos[i][j]; 
      if ( (1-2*ti-tv)<0) {printf("ERROR: log of a negative value in the distance correction process (Kimura 2-param model)\n");exit(-1);}
      /*printf("ti = %.2f     tv = %.2f\n",ti,tv);
      printf("log(1-2*ti-tv) = %.2f\n",log(1-2*ti-tv));
      printf("log(1-2*tv) = %.2f\n",log(1-2*tv));  */
      if (Poids_Transv==0.0) {printf("ERROR: weight of transvertions cannot be 0 in the Kimura 2-param model!!!\n");exit(-1);}
      dcor=2.0*((Poids_Transv+2.0)/Poids_Transv)* 
       ((-0.25)*log(1.0-2.0*ti-tv) +(1.0/8.0)*log(1.0-2.0*tv) );
      Matrice_Dist[i][j]=dcor;
    }
}
/*--------------------------------------------------*/
void Kimura_Correction2()
/* Applies the Kimura 2-parameter model with the formula:
   d_kim = -1/2ln(1-2s/n-v/n)racine(1-2v/n)
   v = nb of transVertions
   s = nb of transItions
   n = nb of characters */
/* "log" here is the natural logarithm : ln */
{
int i,j;
float ti,tv,dcor;
for (i=1;i<=Nb_taxa;i++)
  for (j=1;j<i;j++)
    { ti=(float)TransI[i][j]/NbPos[i][j]; 
      tv=(float)TransV[i][j]/NbPos[i][j]; 
      if ( (1-2*ti-tv)<0) {printf("ERROR: log of a negative value in the distance correction process (Kimura 2-param model)\n");exit(-1);}
      if ( (1.0-2.0*tv)<0) {printf("ERROR: square root of a negative value in the distance correction process (Kimura 2-param model)\n");exit(-1);}
      /* printf("ti = %.2f     tv = %.2f\n",ti,tv);
      printf("(1-2*ti-tv) = %.2f\n",1-2*ti-tv);
      printf("rac(1-2*tv) = %.2f\n",sqrt(1-2*tv));  
      printf("log(A*B) = %.2f\n",log((1-2*ti-tv)*(sqrt(1-2*tv))));*/
      dcor=-0.5*log((1.0-2.0*ti-tv)*sqrt(1.0-2.0*tv));
      /*printf("-0.5*log(A*B)= %.2f\n",-0.5*log((1-2*ti-tv)*(sqrt(1-2*tv))));*/
      Matrice_Dist[i][j]=dcor;
    }
}

/*--------------------------------------------------*/
int main(argc, argv)
int argc;char *argv[];
{extern char *optarg;
char c;
int i,j,corrige=0,liredist=0,sauverDist=0;

/* Traite arguments eventuels */
while ((c = getopt(argc, argv, "d:s:w:c:g:l:")) != -1)
     switch (c) {
     case 'd':liredist=atoi(optarg);break;
     case 's':sauverDist=atoi(optarg);break;
     case 'w':Poids_Transv=atof(optarg);break;
     case 'c':corrige=atoi(optarg);break;
     case 'g':use_gap_pos=atoi(optarg);break;
     case 'l':lower_triang=atoi(optarg);break;
     case '?':{printf("Wrong parameter!!!\n");exit(1);}
                 }
printf("Weight of TransVertions/TransItions = %.2f\n",Poids_Transv);
/*printf("sauverdist : %d\n",sauverDist);
printf("liredist :  %d\n",liredist);
printf("corrige :  %d\n",corrige);*/

if (liredist)
  {Lit_Distances();  /* Affiche_Dist();*/}
else 
  {Lit_Matrice_Car();Calcule_Distances();
  /* corr. if asked and if input=seqs*/
  if (corrige==1) JukesCantor_Correction();
  if (corrige==2) Kimura_Correction2();}

Affiche_Dist();
if (sauverDist) Sauve_Distances();
Infere_quads();
}
