/**********************************************************/
/* Depuis une matrice de caract\`eres considere tous les 
   quadruplets possibles et on tranche pour chacun */

/* ATTENTION : When computing which quartet is supported by a character 
   for a given set of four taxa (t1,t2,t3,t4), only informative characters 
   for parsimony are used (ie, those for which two states appear, both 
   present twice; eg: t1=A t2=A t3=C t4=C support the quartet "t1,t2 | t3,t4").
   Thus, different quartets are usually obtained from the distquart program
   which computes inter-taxa distances using more characters (eg, 
   t1=A t2=A t3=G t4=C). */

/* (Undocumented: pour chq quad especes, on utilise la methode 1/sqrt(pi)
   basee sur les frequences de CARACTERES observes
   dans la matrice M0 de depart. Elle etablie si on peut 
   "SIGNIFICATIVEMENT" decider que la topologie induite par
   le type de car le plus frequent ds Mo est la bonne.     
   Pour cela elle se sert de phenomene sur la LOI du MAX
   concernant les 2 frequences les plus faibles.)            */
/**********************************************************/
/* 
PARAMETERS : 
  -w : weight (real) of transVertions wrt transItions (default is 1.0).
  -g : (1) use gap position whenever is possible or (0) remove them from 
         the analysis (default is 0). 
  -t : cte used in the computation of "significantly less" (undocumented) 
       (default: 0.546)

/* INPUT : file "infile.nuc" nucleotide matrix
   OUTPUT : file "quartfile"  : computed quartets */
/**********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Max_taxa 50
#define Max_car 10000

float Poids_TransV=1.0;
int nbcars,Nb_taxa;
char mat[Max_car+1][Max_taxa+1]; /* mat[i][0] unused */
char Quadr[50],name[Max_taxa+1][11];
float f1,f2,f3,Fsup,Fmoy,Finf;
int use_gap_pos=0; /* default is that gap position are unused */
char gap_position[Max_car];

/*--------------------------------------*/
void Lit_Matrice_cars()
{
FILE *fmat;
int c,e,i;
char car,buffer[255];

if (!(fmat=fopen("infile.nuc","r"))) 
    {printf("ERROR: FILE infile.nuc NOT FOUND\n");exit(-1);}
fgets(buffer,255,fmat);
sscanf(buffer," %d %d",&Nb_taxa,&nbcars);
printf("Input size : %d taxa\nSequences of %d characters\n",Nb_taxa,nbcars);

/* initially no position has a gap */
for (c=0;c<nbcars;c++) gap_position[c]=0;
/* reading the characters */
for (e=1;e<=Nb_taxa;e++)
  {
   for (i=0;i<10;i++) fscanf(fmat,"%c",&name[e][i]); name[e][10]='\0';
   /*printf("species %d = '%s'\n",e,name[e]);*/
   for (c=0;c<nbcars;c++)
     { 
       fscanf(fmat,"%c",&car);
       switch (car) {
       case 'A':mat[c][e]='A';break;	 
       case 'a':mat[c][e]='A';break;
       case 'G':mat[c][e]='G';break;	
       case 'g':mat[c][e]='G';break;	
       case 'C':mat[c][e]='C';break;	
       case 'c':mat[c][e]='C';break;	
       case 'T':mat[c][e]='T';break;
       case 't':mat[c][e]='T';break;
       case 'U':mat[c][e]='T';break; 
       case 'u':mat[c][e]='T';break;
       case '?':mat[c][e]='?';break;
       case '-':{mat[c][e]='-';
		 gap_position[c]++;}break;
       case '.':{if (e==1) 
                  {printf("ERROR: '.' character in the 1st sequence!!!\n");
		   exit(-1);}
	        else mat[c][e]=mat[c][1];};break;
       default: {printf("Warning: unknown character '%c' - treated as a gap\n",car); mat[c][e]='-';gap_position[c]++;} /* si espace ou autre caractere n'en tient pas compte */
       }
     }
   fscanf(fmat,"\n");
  }
fclose(fmat);

/*printf("Set of sequences read :\n");
for (e=1;e<=Nb_taxa;e++) 
  { for (c=0;c<nbcars;c++)
       printf("%c",mat[c][e]);
    printf("\n");}*/
}
/*--------------------------------*/
int Transversion(c1,c2)
char c1,c2;
{
int res=0;
if (( (c1=='A') || (c1=='G') ) && ( (c2=='C') || (c2=='T') )) res=1;
if (( (c1=='C') || (c1=='T') ) && ( (c2=='A') || (c2=='G') )) res=1;
/*if (res) printf("%c <-> %c = transV\n",c1,c2);
else printf("%c <-> %c = transI\n",c1,c2);*/
return(res);
}
/*--------------------------------*/
void Calcule_frequences(e1,e2,e3,e4)
int e1,e2,e3,e4;
{
int c,transV;
float nb1=0.0,nb2=0.0,nb3=0.0,nbtot=0.0;
/* nbtot sert a diviser par le bon nombre de cars qd on donne un
   poids different aux transi et transv */

/* Couting the nb of chars favorizing each of the 3 topology resolution */
f1=0;f2=0;f3=0;
/* f1 pr 12||34, f2 pr 13||24, f3 pr 14||23 */
for (c=0;c<nbcars;c++)
  {
   if (  (mat[c][e1]!='?') && (mat[c][e2]!='?') &&
         (mat[c][e3]!='?') && (mat[c][e4]!='?') )
     if ((use_gap_pos) || (!gap_position[c]))
       if ( (mat[c][e1]!='-') && (mat[c][e2]!='-') && 
            (mat[c][e3]!='-') && (mat[c][e4]!='-') )
        {	
        transV=0;
        if ( (mat[c][e1]!=mat[c][e3]) && 
	     ((mat[c][e1]==mat[c][e2]) && (mat[c][e3]==mat[c][e4]))) 
	 {if (Transversion(mat[c][e1],mat[c][e3]))
           {nb1+=Poids_TransV;transV=1;}
          else {nb1+=1;}
	 }
        if ( (mat[c][e1]!=mat[c][e2]) && 
             (mat[c][e1]==mat[c][e3]) && (mat[c][e2]==mat[c][e4])) 
	 {if (Transversion(mat[c][e1],mat[c][e2]))
           {nb2+=Poids_TransV;transV=1;}
	  else {nb2+=1.0;}
	 }

        if ( (mat[c][e1]!=mat[c][e2]) && 
             (mat[c][e1]==mat[c][e4]) && (mat[c][e2]==mat[c][e3])) 
         {if (Transversion(mat[c][e1],mat[c][e2]))
           {nb3+=Poids_TransV;transV=1;}
          else {nb3+=1.0;}
         }
        if (transV) nbtot+=Poids_TransV; else nbtot+=1.0;
        }
  } /* du for */
if (nbtot)
    {f1=nb1/nbtot;f2=nb2/nbtot;f3=nb3/nbtot;}
/*printf("%d %d %d %d -> f1=%.2f, f2=%.2f, f3=%.2f\n",e1,e2,e3,e4,f1,f2,f3);*/
}
/*----------------------------------*/
int Accepte(Cte)
float Cte;
{
float sigma;
FILE *fecarts;
/* Fsup >= Fmoy >= Finf */
/* Accept if Fsup is statistically significantly greater than Fmoy */
/* ie, if Fsup >= Fmoy +  Cte x sigma x proximity(p2,p3) */
/* where sigma = square-root ( Fmoy x (1-Fmoy) / n ) */
/* (BUT "proximite(Fmoy,Finf)" not yet taken into account) */

sigma = sqrt(Fmoy * (1.0 - Fmoy) / nbcars);
/*printf("sigma = %.4f  Cte = %.4f",sigma,Cte);
printf("  ecart min = %.4f\n",Cte*sigma*1);*/
/* sauve ecartmin et ecartf1f2 */
/*fecarts=fopen("fic_ecarts","a");
fprintf(fecarts,"%.5f %.5f\n",Fsup-Fmoy,Cte*sigma);
fclose(fecarts);*/
if (Fsup > Fmoy + Cte*sigma) return(1);
return(0);
}
/*----------------------------------*/
void Ordonne(e1,e2,e3,e4)
int e1,e2,e3,e4;
{
    /* Ordonne les frequences ds l'ordre decroissant */
if ((f1>=f2) && (f1>=f3)) 
        {
 sprintf(Quadr,"%.2d %.2d || %.2d %.2d\n",e1,e2,e3,e4);
         Fsup=f1;if (f2>=f3) {Fmoy=f2;Finf=f3;} 
                 else {Fmoy=f3;Finf=f2;}
        }
else if ((f2>=f1) && (f2>=f3))
        {
	 sprintf(Quadr,"%.2d %.2d || %.2d %.2d\n",e1,e3,e2,e4);
         Fsup=f2;if (f1>=f3) {Fmoy=f1;Finf=f3;} 
                 else {Fmoy=f3;Finf=f1;}
	}
     else {
           sprintf(Quadr,"%.2d %.2d || %.2d %.2d\n",e1,e4,e2,e3);
           Fsup=f3;
           if (f1>=f2)  {Fmoy=f1;Finf=f2;} 
           else         {Fmoy=f2;Finf=f1;} 
	   }
    
/*printf("les 3 frequences ds l'ordre sont : %.4f , %.4f , %.4f\n",Fsup,Fmoy,Finf);*/
/*printf("j'infere le quadruplet : %s\n",Quadr);*/
}

/*----------------------------------*/
int main(argc, argv)
int argc;
char *argv[];
{ 
extern char *optarg;
int e,i,c;
int nbquads=0,nb=0,nb_gaps=0;
int accept,combien,e1,e2,e3,e4;
float Cte =0.546;
FILE *fquad;
/* RECUPERE LES ARGUMENTS EVENTUELS */
while ((c = getopt(argc, argv, "t:w:g:")) != -1)
     switch (c) {
               case 'w':Poids_TransV=atof(optarg);break;
               case 'g':use_gap_pos=atoi(optarg);break;
               case 't':Cte=atof(optarg);break;/* Cte de signification */
               case '?':{printf("Erreur d'arguments\n");exit(1);}
                }
printf("Weight transV/transI = %.4f\n",Poids_TransV);

Lit_Matrice_cars(); 

for (c=0;c<nbcars;c++) if (gap_position[c]) nb_gaps++;
if (!nb_gaps) printf("No gap positions.\n"); 
else {printf("gaps at %d positions -- these positions ",nb_gaps);
      if (use_gap_pos) printf("are however used whenever is possible\n");
      else printf("are dropped from the whole analysis\n");}

fquad=fopen("quartfile","w");

/* Write nb of taxa */
fprintf(fquad,"   %d\n",Nb_taxa);
/* Indicate the correspondance between taxa nb & taxa name */
for (e=1;e<=Nb_taxa;e++)
  { fprintf(fquad,"%.2d ",e);
    for (i=0;i<10;i++) {fprintf(fquad,"%c",name[e][i]);}
    fprintf(fquad,"\n");
  }
fprintf(fquad,"\n");

/*  Choose the quartet supported by the data for each subset of four taxa */
for (e1=1;e1<=(Nb_taxa-3);e1++)
  for (e2=e1+1;e2<=(Nb_taxa-2);e2++)
    for (e3=e2+1;e3<=(Nb_taxa-1);e3++)
      for (e4=e3+1;e4<=Nb_taxa;e4++)
	{/*printf("(%d,%d,%d,%d)\n",e1,e2,e3,e4);*/nb++;
	 Calcule_frequences(e1,e2,e3,e4);
	 Ordonne(e1,e2,e3,e4); /* set Finf,Fmoy,Fsup */
         accept=Accepte(Cte);
         if ((Fsup!=Fmoy) && (accept)) /* pr le cas 0 */
           { /*printf("%s accepted\n",Arbre_Resolu);*/
            nbquads++;
	    fprintf(fquad,"%s",Quadr);
           }
        } /* du for quad */
printf("%d quartets inferred\n%d quartets leaved unresolved\n",nbquads,nb-nbquads);
printf("Quartets output in file : quartfile\n");
fclose(fquad);
}


