/***************************************************************/
/* Programme qui infere la relation ||* depuis la relation ||.
   C'est-a-dire qu'il infere les bips de l'arbre le plus 
   proche de la relation de voisinage du depart                */
/***************************************************************/
/* ENTREE = Liste de quads ds le fichier "quads.0"             */
/* SORTIE = Liste des bipartitions induites ds "bipfile"       */
/*          "quartfile.res" : quds de Q*                        */
/*          "quartfile.left" : quads de Q-Q*                    */
/* Compx = O(n^4) : algo dirige par les aretes                 */
/* Les ensembles Xj sont geres par listes doublement chainees  */
/***************************************************************/

/* A FAIRE : 

Essayer d'utliser un max la memoire dynamique car il faudrait faire
un version qui marche pour plusieurs centaines d'especes.

OPTIMISATION !!!!!!!!!!!!!!!!  :

Lors du 2eme parcours (descendant prefixe), d\`es qu'on rencontre une
arete e_i pour laquelle l'etiquette calculee lors du 1er parcours est
"oui" (i.e., i peut etre insere au dessus), alors se contenter de
mettre les etiquettes a "non" dans toute la sous-arbo inferieure de
e_i, SANS MEME REGARDER LES QUADRUPLETS DE Q OU METTRE A JOUR LES
STRUCTURES Z (on en aura pas besoin pour un parcours futur). Par contre 
on ne peut pas faire ca lors du 1er parcours car il faut que les struct
Z soient a jour puisque le 2eme parcours en a besoin).

Justification : si i peut etre insere au dessus de e_i, il ne pourra
pas etre insere au-dessous. 

Autre OPTIMISATION : ne pas considerer les especes dans n'importe quel
ordre, par exemple, on peut lors d'un examen de Q compter pour chaque
espece le nb de 4-arbres dans lesquels elle est impliquee (et aussi 
peut-etre avec quelles autres esp elle est le plus souvent voisine).

Ca devrait donner des infos pour savoir un peu dans quel ordre il
vaut mieux inserer les especes. Le mieux est peut-etre d'inserer avant tout
les especes les plus frequentes, de facon a avoir des Qi de taille reduite
quand l'arbre T* commence a avoir plus d'aretes ? et l'inverse ?
Faire des essais, pour savoir ce qui est le plus rapide.

Autre OPTIMISATION : Lors du 2eme parcours, avant de remonter dans
l'arbre (i.e. fin de traitement d'une arete), on peut en
profiter pour supprimer l'arete si elle est "non"x2 !? 

Autre OPTIMISATION : binariser artificiellement l'arbre, car rien
n'est plus couteux que la topologie etoile : on a du O(n^4) pour tout
ce qui est gestion des structures Z, ... alors que si l'arbre est
binaire, on a du O(n^3) pour les memes choses (on n'a plus que du
O(n^4) lie au nombre de 4-arbres examines dans Q).
Toutefois, est-ce que ca amene un GAIN EN TEMPS D'EXECUTION ?
Bien reflechir d'abord en fonction des struct actuelles 

Peut-etre que cette idee ne vaut le coup qu'au dela d'un certain nombre 
d'especes ? faire des essais et parametrer le programme par un simple
test en debut.

*/

#include <stdio.h>
#include <stdlib.h>

#define Max_nb_taxa 50
#define Inf 0
#define Sup 1
#define Prefixe 0
#define Postfixe 1
#define Non 0
#define Oui 1

typedef struct tliste {int taxon; struct tliste *suiv;} tliste;
/* L'arbre T_i est gere par pointeurs, chaque noeud et chaque arete
   etant une entite distincte. L'arbre est enracine a l'espece 1 */

typedef struct tarete    {
                          struct tnoeud *fils,*pere;
			  tliste *X[2]; /* Inf et Sup */
				      /* listes dblmt chainees */
			  int Z[2][Max_nb_taxa+1]; /* on a parfois besoin de
						 Zj des deux sens a la fois */
			  int horsde[2]; /* 0=non 1=oui */
                         } tarete;

typedef struct listarete { tarete *a; struct listarete *suiv;} listarete ;

typedef struct tnoeud    {
                         int taxon; /* num du taxon ou num nd interne */
                         struct tarete *haut;
			 listarete *bas,*fin;  
                         } tnoeud;


/* --------------Var globales-------------- */
int Nb_taxa,nbquads,Num_intern,Etape;
tnoeud *Taxon1; /* la racine de l'arbre */

char name[Max_nb_taxa+1][11];

/* Declaration de la matrice contenant les quadruplets inferes 
   on a toujours i1<i2, i3<i4, i1<i3 */
short MatQuad[Max_nb_taxa+1][Max_nb_taxa+1][Max_nb_taxa+1][Max_nb_taxa+1];
short QuadRes[Max_nb_taxa+1][Max_nb_taxa+1][Max_nb_taxa+1][Max_nb_taxa+1];
short Nb_bips=0,bips[2*(Max_nb_taxa+1)-3][2][Max_nb_taxa+1]; /* taxa pas ranges
							 dans le bon ordre */
/*---------------------------------------------------------------*/
void Affiche_Arbre(n)
tnoeud *n;
{
listarete *Sj;
for (Sj=n->bas;Sj;Sj=Sj->suiv)
  {printf("%d --(%d|%d)-- %d\n",n->taxon,Sj->a->horsde[Sup],Sj->a->horsde[Inf],Sj->a->fils->taxon);
   Affiche_Arbre(Sj->a->fils);}
}

/*-----------------------------------------*/
void Affiche_Xj(Sj,sens)
tarete *Sj;int sens;
{tliste *x;
for (x=Sj->X[sens];x;x=x->suiv) printf(" %d",x->taxon); 
printf(".\n");
}
/*-----------------------------------------*/
void Lit_quadruplets()
{
FILE *fquad;
char neoquad[100];
int t1,t2,t3,t4,tmp;

/* init l'ens de quads */
for (t1=1;t1<=Nb_taxa;t1++)
  for (t2=t1+1;t2<=Nb_taxa;t2++)
    for (t3=t1+1;t3<=Nb_taxa;t3++)
      for (t4=t3+1;t4<=Nb_taxa;t4++)
        MatQuad[t1][t2][t3][t4]=0;
nbquads=0;
if (!(fquad=fopen("quartfile","r"))) {printf("ERROR READING FILE quartfile\n");exit(-1);}

/* Lit le nombre de taxa */
fscanf(fquad,"%d",&Nb_taxa);
printf("%d taxa\n",Nb_taxa);
if (Nb_taxa>50) {printf("Sorry, this program can't handle more than 50 taxa.\n");exit(0);}
/* Lit la correspondance numero-espece */
for (t1=1;t1<=Nb_taxa;t1++)
  fscanf(fquad,"%d %s\n",&tmp,&name[t1]);
fscanf(fquad,"\n");

/*printf("Taxa :\n");
for (t1=1;t1<=Nb_taxa;t1++) printf("%s\n",name[t1]); 
printf("\n");*/

/* Lit les quadruplets */
while (fscanf(fquad,"%s",neoquad)!=EOF)
  {
   /* nouveau quadruplet */
   nbquads++;
   t1=atoi(neoquad);
   fscanf(fquad,"%s",neoquad);
   t2=atoi(neoquad);
   /* ordonne les taxas par ordre croissant */
   if (t1>t2) {tmp=t1;t1=t2;t2=tmp;}
   fscanf(fquad,"%s",neoquad); /* lit le || */
   fscanf(fquad,"%s",neoquad);
   t3=atoi(neoquad);
   fscanf(fquad,"%s",neoquad);
   t4=atoi(neoquad);
   /* ordonne les taxas par ordre croissant */
   if (t3>t4) {tmp=t3;t3=t4;t4=tmp;}
   if (t1<t3) {/*printf("(%d,%d,%d,%d)\n",t1,t2,t3,t4);*/
	       MatQuad[t1][t2][t3][t4]=1;}
   else {/*printf("(%d,%d,%d,%d)\n",t3,t4,t1,t2);*/
	 MatQuad[t3][t4][t1][t2]=1;}
   /*   if (t1<t3) printf("%d%d || %d%d , ",t1,t2,t3,t4);
   else printf("%d%d || %d%d , ",t3,t4,t1,t2); */
  }
printf("%d input quartets\n",nbquads);
}
/*---------------------------------*/
int Appartient_Q(x,y,z,i)
int x,y,z,i;
/* regarde si le 4-arbre xy||zi est dans Q */
{int tmp;
if (x>y) {tmp=x;x=y;y=tmp;} /* on sait z<i */
if (z<x) {tmp=x;x=z;z=tmp; tmp=y;y=i;i=tmp;}
if (MatQuad[x][y][z][i]) return(Oui); else return(Non);
}
/*--------------------------------*/
tliste *Neo_elem(tax)
int tax;
{tliste *t;
t=(tliste *)malloc(sizeof(tliste));
t->taxon=tax;t->suiv=NULL;return(t);
}
/*---------------------------------*/
tliste *Union_Ensembles_X(Xl,Xm)
tliste *Xl,*Xm;
/* SE : Union d'ens disjoints en O(n) car il faut RECOPIER LA LISTE Xm */
{tliste *neo,*x;
for (x=Xm;x;x=x->suiv) 
  {
   neo=(tliste *)malloc(sizeof(tliste));
   if (!neo) {printf("ALLOCATION FOIRE !!!\n");exit(-1);}
   neo->taxon=x->taxon;
   neo->suiv=Xl;
   Xl=neo;
  }
return(Xl);
}

/*---------------------------------*/
void New_fils(nd,typ,num,ou)
tnoeud *nd;char typ;int num;char ou; 
{/* Ajoute un fils de 'typ' interne ('I') ou externe ('E') au noeud
    'nd'. Ce neo noeud est numerote 'num' et est rajoute au debut
    (ou='d') ou a la fin (ou='f') des fils de nd. On cree aussi
    l'arete entre les 2 noeuds. */
int obj;listarete *neo;
if (nd==NULL) {printf("erreur noeur pere inexistant\n");exit(-1);}
/*printf("nd =%d, typ = %c, num = %d\n",nd->taxon,typ,num);*/
/* cree l'arete */
neo=(listarete*)malloc(sizeof(listarete));
neo->a=(tarete*)malloc(sizeof(tarete));
neo->a->pere = nd;
neo->a->horsde[Inf]=neo->a->horsde[Sup]=-1;
for (obj=1;obj<Etape;obj++) 
  neo->a->Z[Inf][obj] = neo->a->Z[Sup][obj] = 0;
neo->a->X[Sup]=NULL;
if (typ=='E') neo->a->X[Inf]=Neo_elem(num);
else          neo->a->X[Inf]=NULL;
/* cree le noeud fils */
neo->a->fils=(tnoeud*)malloc(sizeof(tnoeud));
neo->a->fils->haut = neo->a ;
neo->a->fils->taxon=num;
if (typ=='I') Num_intern++;  /* si interne incremente le cptr */
neo->a->fils->bas = neo->a->fils->fin = NULL;

if (ou=='f')
  {/* rattache le tout a 'nd' en fin */
    neo->suiv=NULL;
    if (!nd->bas) nd->bas = neo; /* si 1er fils */
    else nd->fin->suiv=neo; 
    nd->fin = neo;
  }
else 
  {/* rattache le tout a 'nd' en debut */
    if (!nd->bas)  /* si 1er fils */
         {neo->suiv=NULL; nd->bas = nd->fin = neo;}
    else {neo->suiv=nd->bas; nd->bas = neo;}
  }
}
/*---------------------------------*/
void Init_Arbre()
{
tnoeud *n;
/* Init l'arbre T_i depuis la resolution du quadruplet (1,2,3,4) */
/* Creer manuellement le taxon 1 sert de racine a l'arbre */
Taxon1=(tnoeud*)malloc(sizeof(tnoeud));
Taxon1->taxon=1;Taxon1->haut=NULL;
New_fils(Taxon1,'I',Num_intern,'f'); /* garder 'I' car init faites */
/* Cas particulier : la racine est au-dessus */
Taxon1->bas->a->X[Sup]=Neo_elem(1);
n = Taxon1->bas->a->fils ; /* on y descend */
if (MatQuad[1][2][3][4]) 
  {
    New_fils(n,'E',2,'f'); 
    New_fils(n,'I',Num_intern,'f');
    n = n->fin->a->fils ; /* on y descend */
    New_fils(n,'E',3,'f');New_fils(n,'E',4,'f');
  }
else if (MatQuad[1][3][2][4])
  {
    New_fils(n,'E',3,'f'); 
    New_fils(n,'I',Num_intern,'f');
    n = n->fin->a->fils ; /* on y descend */
    New_fils(n,'E',2,'f');New_fils(n,'E',4,'f');
  }
else if (MatQuad[1][4][2][3])
  {
    New_fils(n,'E',4,'f'); 
    New_fils(n,'I',Num_intern,'f');
    n = n->fin->a->fils ; /* on y descend */
    New_fils(n,'E',2,'f');New_fils(n,'E',3,'f');
  }
else /* irresolution */
  { New_fils(n,'E',2,'f'); New_fils(n,'E',3,'f'); New_fils(n,'E',4,'f'); }
}
/*----------------------------*/
int Vide(Z)
int Z[];
{int x;
/*printf("Vide ? ");*/
for (x=1;x<Etape;x++)
  if (Z[x]) {/*printf("Non\n");*/return(Non);}
/*printf("Oui\n");*/
return(Oui);
}
/*-----------------------------*/
int Hors_de(Sj,sens)
tarete *Sj;int sens;
{/* Le traitement Sup ou Inf est le meme mais doit etre distingue a cause
    de l'orientation de l'arbre et des structures de donnees (bas!=haut..) */
int obj;
listarete *Sl,*Sm,*frere;
tliste *x,*y,*z;
/* Si on a 1 seul elem du cote du parcours, renvoyer ok tout de suite */
if ((sens==Inf) && (!Sj->fils->bas)) return(Oui); /* les Zj restent vides */
if ((sens==Sup) && (!Sj->pere->haut)) return(Oui);
/*------------*/
/* Calcul Z_j */
/*------------*/
if (sens==Inf)
  {
    /* Init le Zj de l'edi Sj par union depuis ses fils */
    for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
      for (obj=1;obj<Etape;obj++) 
	if (Sl->a->Z[Inf][obj]) Sj->Z[Inf][obj]=1;
  }
else /* sens = Sup */
  { 
    /* Init le Zj de l'edi Sj par union du Sup du pere et des Inf des freres */
    for (obj=1;obj<Etape;obj++) 
      if (Sj->pere->haut->Z[Sup][obj]) Sj->Z[Sup][obj]=1;
    for (frere=Sj->pere->bas;frere;frere=frere->suiv)
      if (frere->a!=Sj) /* sauf lui-meme */
	for (obj=1;obj<Etape;obj++)
	  if (frere->a->Z[Inf][obj]) Sj->Z[Sup][obj]=1;
  }
/* Retire de Zj les objets de Xj */
for (x=Sj->X[sens];x;x=x->suiv)
  Sj->Z[sens][x->taxon] = 0;

/*printf("Zj =");
for (obj=1;obj<Etape;obj++) if (Sj->Z[sens][obj]) printf(" %d",obj);
printf(".\n");
printf("Xj =");
for (x=Sj->X[sens];x;x=x->suiv) printf(" %d",x->taxon);
printf("\n");*/
/*----------------------*/
/* Examen des I_j^{l,m} */
/*----------------------*/
/* Pour chq couple d'edi-fils Sl et Sm de l'edi courant Sj */
if (sens==Inf)
  {/*printf("calcul non-trivial\n");*/
  for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
   for (Sm=Sl->suiv;Sm;Sm=Sm->suiv)
     { 
       for(x=Sl->a->X[sens];x;x=x->suiv)
         {/*printf("x=%d\n",x->taxon);*/
	 for(y=Sm->a->X[sens];y;y=y->suiv)
	    {/*printf(" y=%d\n",y->taxon);*/
	    for(z=Sj->X[!sens];z;z=z->suiv)
	      {/*printf("  z=%d\n",z->taxon);*/
	      if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		Sj->Z[sens][z->taxon]=1;
	      }
	    }
	 }
     }
 }
else /* sens = Sup */
  {/*printf("calcul non-trivial : (pere,frere)\n");*/
     /* Pour chq couple (Pere , edi-fils Sl) */
    for (Sl=Sj->pere->bas;Sl;Sl=Sl->suiv)
      if (Sl->a!=Sj)
	for(x=Sj->pere->haut->X[sens];x;x=x->suiv)
	  {/*printf("x=%d\n",x->taxon);*/
	  for(y=Sl->a->X[!sens];y;y=y->suiv)
	    {/*printf(" y=%d\n",y->taxon);*/
	    for(z=Sj->X[!sens];z;z=z->suiv)
	      {/*printf("  z=%d\n",z->taxon);*/
	      if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		Sj->Z[sens][z->taxon]=1;
	      }
	    }
	  }
   /* Pour chq couple (edi-fils Sl , edi-fils Sm*/
    /*printf("suite : (frere,frere)\n");*/
    for (Sl=Sj->pere->bas;Sl;Sl=Sl->suiv)
      if (Sl->a!=Sj)
	for (Sm=Sl->suiv;Sm;Sm=Sm->suiv)
	  if (Sm->a!=Sj)
	    for(x=Sl->a->X[!sens];x;x=x->suiv)
	      {/*printf("x=%d ",x->taxon);*/
	      for(y=Sm->a->X[!sens];y;y=y->suiv)
		{/*printf(" y=%d ",y->taxon);*/
		for(z=Sj->X[!sens];z;z=z->suiv)
		  {/*printf("  z=%d\n",z->taxon);*/
		  if (!Appartient_Q(x->taxon,y->taxon,z->taxon,Etape)) 
		    Sj->Z[sens][z->taxon]=1;
		  }
		}
	      }
  }
if (Vide(Sj->Z[sens])) return(Oui); else return(Non);
}
/*-----------------------------*/
void Calcule_les_Xj(Sj,sens)
tarete *Sj;int sens;
{listarete *Sl,*frere;tliste *x,*pr;
/* Si non trivial remettre Xj a zero */
if ( ((sens==Inf) && (Sj->fils->bas)) || 
     ((sens==Sup) && (Sj->pere->haut)) )
  if (Sj->X[sens])   /* liberer de l'etape precedente */ 
    { /*printf("non trivial ");*/
      /*pr=Sj->X[sens]; for(x=pr->suiv;x;x=x->suiv) free(pr);
      free(pr);*/ 
      Sj->X[sens]=NULL;}

if (sens==Inf) /* faire un parcours Postfixe */
  {
    for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
      { 
	Calcule_les_Xj(Sl->a,sens); /* Parcours les fils */
	/* Calcul le Xj de l'edi Sj par union depuis ses fils */
	Sj->X[Inf]=Union_Ensembles_X(Sj->X[Inf],Sl->a->X[Inf]);
      }
    /*printf("resultat : ");Affiche_Xj(Sj,sens);*/
  }
else /* pr les Sup faire un parcours prefixe */
  {    
   if (Sj->pere!=Taxon1) /* Si racine : Xj dj positionne */
     {
      /* Calcul le Xj depuis l'union du Sup du pere et des Inf des freres */
      Sj->X[Sup]=Union_Ensembles_X(Sj->X[Sup],Sj->pere->haut->X[Sup]);
      for (frere=Sj->pere->bas;frere;frere=frere->suiv)
	{
	 if (frere->a!=Sj) /* sauf lui-meme */
	   Sj->X[Sup]=Union_Ensembles_X(Sj->X[Sup],frere->a->X[Inf]);
	 /*printf("Apres l'union avec le frere Xj = ");Affiche_Xj(Sj,sens);*/
       }
     }
   /* Parcours les fils */
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv)
     Calcule_les_Xj(Sl->a,sens);
  }

}
/*-----------------------------*/
void Parcours_Prof(Sj,sens,facon)
tarete *Sj;int sens,facon;
{listarete *Sl;
if (facon==Postfixe)
  {
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
     Parcours_Prof(Sl->a,sens,facon);
   Sj->horsde[sens] = Hors_de(Sj,sens);}
else
  {
   Sj->horsde[sens] = Hors_de(Sj,sens);
   for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) 
     Parcours_Prof(Sl->a,sens,facon);}
}
/*-----------------------------*/
void Enlever_double_non(n)
tnoeud *n;
/* efface les aretes etiquetees par "non" dans les deux sens et
leur noeud inferieur. Rapatrie ses aretes filles en fin des aretes
filles du noeud superieur en O(1) */
{listarete *Sl,*Sm,*prec;
tliste *x,*pr;int prem=Non;
Sl=n->bas;prec=n->bas;
while (Sl)
  {
    Enlever_double_non(Sl->a->fils); /* traiter d'abord la sous-arbo */
    if ((Sl->a->horsde[Inf]==Non) && (Sl->a->horsde[Sup]==Non))
      { /* recupere la sous-arbo de l'arete eliminee */
	for (Sm=Sl->a->fils->bas;Sm;Sm=Sm->suiv) Sm->a->pere=n;
	if (Sl->suiv) Sl->a->fils->fin->suiv=Sl->suiv;
	else n->fin=Sl->a->fils->fin; /* il pointe dj sur rien */
	if(prec!=Sl) prec->suiv=Sl->a->fils->bas; 
	else {prem=Oui;prec=n->bas=Sl->a->fils->bas;}
       /* detruit tout ce qui touche au noeud et arete collapsed */
	/* free(Sl->a->fils);
	pr=Sl->a->X[Inf]; if (pr) 
	  { for(x=pr->suiv;x;x=x->suiv) free(pr);  free(pr);}
	pr=Sl->a->X[Sup]; if (pr) 
	  { for(x=pr->suiv;x;x=x->suiv) free(pr);  free(pr);}
	free(Sl->a);free(Sl); */
	/* et ca repart !*/
	if (prem) Sl=prec; else Sl=prec->suiv; 
      }
    else {prec=Sl;Sl=Sl->suiv;} /* ok meme si premier elem */
  }
}
/*-----------------------------*/
void Sur_arete(nd,i)
tnoeud *nd;int i;
{ listarete *Sl;
/*printf("Insere sur arete au dessus de %d\n",nd->taxon);*/
if (nd->bas) /* on est sur un noeud interne */
  {
    /* se cree un nouveau 1er fils */
    New_fils(nd,'I',Num_intern,'d');
    /* lui fait passer toute la sous-arbo d'avant */
    for(Sl=nd->bas->suiv;Sl;Sl=Sl->suiv) Sl->a->pere=nd->bas->a->fils;
    nd->bas->a->fils->bas=nd->bas->suiv;
    nd->bas->a->fils->fin=nd->fin;
    nd->bas->suiv=NULL; nd->fin = nd->bas;
  }
else /* on est sur un noeud externe */
  { /* on transfert le taxon en 1 neo noeud fils */
    New_fils(nd,'E',nd->taxon,'d');
    nd->taxon=Num_intern;Num_intern++;
  }
/* rajoute en 2eme fils le neo taxon */
New_fils(nd,'E',i,'f');    
}
/*-----------------------------*/
void Sur_noeud(nd,i)
tnoeud *nd;int i;
{/*printf("Insere sur noeud %d\n",nd->taxon);*/
New_fils(nd,'E',i,'f');
}
/*-----------------------------*/
void Insere_tax_arbre(n,i,ok)
tnoeud *n;int i,*ok;
/* deduit l'endroit ou greffer le taxon i depuis les etiquettes des aretes
   de l'arbre. `ok' sert a arreter la descente en prof (on depile juste) */
{listarete *Sl;
for (Sl=n->bas;(Sl) && (*ok==Non);Sl=Sl->suiv)
  {
    if (Sl->a->horsde[Sup]==Oui)
      if (Sl->a->horsde[Inf]==Oui) {Sur_arete(Sl->a->fils,i); *ok=Oui;}
      else 
	if (Sl->a->fils->bas) Insere_tax_arbre(Sl->a->fils,i,ok);
  }
 if (!(*ok)) /* tout indique le noeud courant */ 
   {Sur_noeud(n,i);*ok=Oui;}
}
/*-----------------------------*/
void Raz_Zj(Sj)
tarete *Sj;
{int obj;listarete *Sl;
for (obj=1;obj<Etape;obj++) {Sj->Z[Inf][obj]=0;Sj->Z[Sup][obj]=0;}
for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) Raz_Zj(Sl->a);
}
/*-----------------------------*/
void Inclus_neo_tax(i)
int i;
{int ok=Non;
/*printf("Calcule les Xj des edi Inf\n");*/
Calcule_les_Xj(Taxon1->bas->a,Inf);
/*printf("calcule les Xj des edi Sup (forcement apres les Inf)\n");*/
Calcule_les_Xj(Taxon1->bas->a,Sup);

/* necessaire ? */
/* Si oui peut etre combine avec Calcul des Xj ci-dessus */
/* Si on l'enleve on a pas les memes resultats dans certains cas !!! */ 
Raz_Zj(Taxon1->bas->a);

/*printf("parcours 1 en prof pr les etiquettes\n");*/
Parcours_Prof(Taxon1->bas->a,Inf,Postfixe);
/*printf("parcours 2 en prof pr les etiquettes\n");*/
Parcours_Prof(Taxon1->bas->a,Sup,Prefixe);
/*printf("Apres avoir teste insertion de l'objet %d\n",i);
Affiche_Arbre(Taxon1);*/
Enlever_double_non(Taxon1);
/*printf("Apres avoir enleve les aretes non-supportees :\n");
Affiche_Arbre(Taxon1);*/
Insere_tax_arbre(Taxon1,i,&ok);
/*printf("Apres avoir insere le nouvel objet:\n");
Affiche_Arbre(Taxon1);*/
}
/*--------------------------*/
void Sauve_Bips()
{
FILE *f;
int b,t;
f=fopen("bipfile","w");
fprintf(f,"   %d\n",Nb_taxa);
for (t=1;t<=Nb_taxa;t++)
  fprintf(f,"%.2d %s\n",t,name[t]);
fprintf(f,"\n");
for (b=0;b<Nb_bips;b++)
  {
    for(t=0;bips[b][0][t];t++)  fprintf(f,"%.2d ",bips[b][0][t]);
    fprintf(f,"|");
    for(t=0;bips[b][1][t];t++)  fprintf(f," %.2d",bips[b][1][t]);
    fprintf(f," (32000)\n");
  }
fclose(f);
printf("Bipartitions (edges) of the tree output in the file : bipfile\n");
}
/*--------------------------*/
void Calcule_Bipartitions(Sj)
tarete *Sj;
{listarete *Sl;tliste *x;
int taxons[Max_nb_taxa+1],t,ind; /* pr les sauver dans l'ordre */
for (t=1;t<=Nb_taxa;t++) bips[Nb_bips][0][t]=bips[Nb_bips][1][t]=0;
for (x=Sj->X[Inf],ind=0;x;x=x->suiv,ind++)  
  bips[Nb_bips][0][ind]=x->taxon;
bips[Nb_bips][0][ind]=0;
for (x=Sj->X[Sup],ind=0;x;x=x->suiv,ind++)  
  bips[Nb_bips][1][ind]=x->taxon;
bips[Nb_bips][1][ind]=0;
Nb_bips++;
for (Sl=Sj->fils->bas;Sl;Sl=Sl->suiv) Calcule_Bipartitions(Sl->a);
}

/*-----------------------------*/
void Etablit_Bipartitions()
{/* meme les bips triviales (emboite en a besoin) */
/* besoin de recalculer les X a nouveaux ? */
Calcule_les_Xj(Taxon1->bas->a,Inf);
Calcule_les_Xj(Taxon1->bas->a,Sup);
Calcule_Bipartitions(Taxon1->bas->a);
}

/*-----------------------------*/
void Sauve_Quads_non_Resolus()
{
/* met les quadruplets resolus dans "quartfile.res" 
et les autres dans "quartfile.left" */
int b,g1,g2,d1,d2,t1,t2,t3,t4,tmp,nblus;
FILE *fr,*f1;
nblus=nbquads;
/* init */
for (t1=1;t1<=Nb_taxa;t1++)
  for(t2=t1+1;t2<=Nb_taxa;t2++)
    for (t3=1;t3<=Nb_taxa;t3++)
      for(t4=t3+1;t4<=Nb_taxa;t4++)
	QuadRes[t1][t2][t3][t4]=0;
/* met ceux impliques par les bips dans QuadRes */
for (b=0;b<Nb_bips;b++)
  if ((bips[b][0][1]) && (bips[b][1][1])) /* si bip non triviale */
  {  
  for (g1=0;bips[b][0][g1];g1++)
    for (g2=g1+1;bips[b][0][g2];g2++)
      {
	/* remet a l'endroit */
	t1=bips[b][0][g1];t2=bips[b][0][g2];
	if (t1>t2) {tmp=t1;t1=t2;t2=tmp;}
	for (d1=0;bips[b][1][d1];d1++)
	  for (d2=d1+1;bips[b][1][d2];d2++)
	    {
	      /* le remet a l'endroit */
	      t3=bips[b][1][d1];t4=bips[b][1][d2];
	      if (t3>t4) {tmp=t3;t3=t4;t4=tmp;} 
	      if (t1<t3) 
		{/*printf("(%d,%d,%d,%d) ",t1,t2,t3,t4);*/
		 QuadRes[t1][t2][t3][t4]=1;
		 if (MatQuad[t1][t2][t3][t4]) nbquads--;
		 MatQuad[t1][t2][t3][t4]=0;}
	      else 
		{/*printf("(%d,%d,%d,%d)\n",t3,t4,t1,t2);*/
		 QuadRes[t3][t4][t1][t2]=1;
		 if (MatQuad[t3][t4][t1][t2]) nbquads--;
		 MatQuad[t3][t4][t1][t2]=0;}
	    }
      }
   }
printf("%d quartets in Q* (output in file : quartfile.res)\n",nblus-nbquads);
printf("%d quartets left in Q (output in file : quartfile.left)\n",nbquads);
printf("Treelikeness of the data set = %.2f%%\n",(float)(nblus-nbquads)*100.0/(float)nblus);

fr=fopen("quartfile.res","w");f1=fopen("quartfile.left","w");
fprintf(fr,"   %d\n",Nb_taxa);fprintf(f1,"   %d\n",Nb_taxa);
for (t1=1;t1<=Nb_taxa;t1++)
  {fprintf(fr,"%.2d %s\n",t1,name[t1]);fprintf(f1,"%.2d %s\n",t1,name[t1]);}
fprintf(fr,"\n");fprintf(f1,"\n");
for (t1=1;t1<=Nb_taxa;t1++)
  for(t2=t1+1;t2<=Nb_taxa;t2++)
    for (t3=t1+1;t3<=Nb_taxa;t3++)
      for(t4=t3+1;t4<=Nb_taxa;t4++)
	{if (QuadRes[t1][t2][t3][t4]) 
	   fprintf(fr,"%.2d %.2d || %.2d %.2d\n",t1,t2,t3,t4);
	 if (MatQuad[t1][t2][t3][t4]) 
	   fprintf(f1,"%.2d %.2d || %.2d %.2d\n",t1,t2,t3,t4);}
fclose(f1);fclose(fr);
}

/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/        

int main(argc, argv)
int argc;char *argv[];
{extern char *optarg;
char c;
/* Nb taxa par defaut */
Nb_taxa=4;
/* Lit l'ensemble Q depuis fic */
Lit_quadruplets();
if (nbquads==0) printf("ERROR ? quartfile is EMPTY !!!!\n");
/* T_4 = resolution du quadruplet (1,2,3,4) */
Etape=4;Num_intern=Max_nb_taxa+1;
Init_Arbre();
/*Affiche_Arbre(Taxon1);*/
printf("\nIncorporating taxa in the tree :\n....");
/* Insere une nouvelle espce a chaque etape */
for (Etape=5;Etape<=Nb_taxa;Etape++)
    {Inclus_neo_tax(Etape);printf(".");fflush(stdout);}
printf("\n");
/* Sauve les bipartitions B*  */
Etablit_Bipartitions();
Sauve_Bips();
Sauve_Quads_non_Resolus();
}

